package Aplpl::ENV;
use strict;

use Aplpl::VALUE;
use Aplpl::PAIR;
use Aplpl::EXP;
use Aplpl::NAME;
use Exporter;
use vars qw(@ISA @EXPORT $DEFINE $LAMBDA $PINCR $PADD $PSUB $PMLT 
            $PDIV $PPRINT $PLOAD $PEQUAL $PGE $PLT $PQUIT $eof 
            $falseValue $trueValue $trueSym);
@ISA = ("Exporter");
@EXPORT = qw(&newENV &initGlobalEnv &bindingInFrame &bindingInEnv 
             &defVar &setVar &getVar &extendEnv &isbound  &printEnv 
             $eof $falseValue $trueValue $trueSym $DEFINE $LAMBDA 
             $PINCR $PADD $PSUB $PMLT $PDIV $PPRINT $PLOAD $PEQUAL 
             $PGE $PLT $PQUIT);

$DEFINE = defName("define");
$LAMBDA = defName("lambda");
$PINCR  = defName("incr");
$PADD   = defName("+");
$PSUB   = defName("-");
$PMLT   = defName("*");
$PDIV   = defName("/");
$PPRINT = defName("print");
$PLOAD  = defName("load");
$PEQUAL = defName("equal");
$PGE    = defName(">=");
$PLT    = defName("<");
$PQUIT  = defName("quit");

$eof        = newEXP ('VXP', newVAL ('STR', "*eof*"));
$falseValue = newVAL ('SYM', defName("false"));
$trueValue  = newVAL ('SYM', defName("true"));
$trueSym    = newEXP ('RXP', newVAL ('SYM', defName("true")));

sub initGlobalEnv {
  my $op;

  my $globalEnv = emptyENV();
  for ($op = $PINCR; $op <= $PQUIT; $op++) {
     defVar ($op, newVAL('PRIM', $op), $globalEnv);
  }
  my $setName = defName ("vmzero");
  my $setVal = newVAL('INT', 0);
  defVar ($setName, $setVal, $globalEnv);

  $setName = defName ("vmtrue");
  $setVal = $trueValue;
  defVar ($setName, $setVal, $globalEnv);

  $setName = defName ("vmfalse");
  $setVal = $falseValue;
  defVar ($setName, $setVal, $globalEnv);

  return $globalEnv;
}

sub emptyENV {
   my $type  = 'ENV';
   my $vars  = $nil;
   my $vals  = $nil;
   my $next  = $null;

   my $objref     = sub {
                         my $msg = shift;
                         if ($msg eq 'getvars') {
                            return $vars;
                         } 
                         elsif ($msg eq 'getvals') {
                            return $vals;
                         } 
                         elsif ($msg eq 'setvars') {
                            my $lvars = shift;
                            $vars = $lvars;
                            return 1;
                         } 
                         elsif ($msg eq 'setvals') {
                            my $lvals = shift;
                            $vals = $lvals;
                            return 1;
                         } 
                         elsif ($msg eq 'getnext') {
                            return $next;
                         } 
                         elsif ($msg eq 'isempty') {
                            return ($vars == $nil && $next eq $null);
                         } 
                         elsif ($msg eq 'type') {
                            return $type;
                         } 
                         else {
                            $main::ots = $main::ots .  sprintf("%s", 
                                   "emptyENV: message unknown: $msg !\n");
                            die "Abnormal Termination.";
                         }
                    };
   return (bless $objref);
}

sub bindingInFrame {
   my $nl    = shift;
   my $vl    = shift;
   my $nm    = shift;
   my $lfound = 0;

   while ($nl != $nil && !$lfound) {
      if (car($nl) == $nm) {
         $lfound = 1;
      }
      else {
         $nl = cdr($nl);
         $vl = cdr($vl);
      }
   }
   return $vl; 
}

sub printEnv {
   my $menv = shift;

   if ($menv->('isempty')) {
      print "empty environment<br>";
   }
   else {
      prNL($menv->('getvars'));
      $menv   = $menv->('getnext');   
      while ($menv ne $null) {
         prNL($menv->('getvars'));
         $menv  = $menv->('getnext');   
      }
   }
}


sub bindingInEnv {
   my $nm   = shift;
   my $lenv = shift;
   my $vl   = bindingInFrame($lenv->('getvars'), 
                             $lenv->('getvals'), 
                             $nm);
   if ($vl == $nil) {
      if ($lenv->('isempty')) {
         return $nil;
      }
      else {
         $lenv    = $lenv->('getnext');   
         while ($vl == $nil && $lenv ne $null) {
            $vl    = bindingInFrame($lenv->('getvars'), 
                                    $lenv->('getvals'), 
                                    $nm);
            $lenv  = $lenv->('getnext');   
         }
         return $vl;
      }
   }
   else {
      return $vl;
   }
};

sub defVar  {
   my $nm   = shift;
   my $val  = shift;
   my $lenv = shift;
  
   $lenv->('setvars', nlcons($nm, $lenv->('getvars'))); 
   $lenv->('setvals', vlcons($val, $lenv->('getvals'))); 
};

sub setVar  {
   my $nm   = shift;
   my $val  = shift;
   my $lenv = shift;
   my $vl   = bindingInEnv($nm, $lenv);

   if ($vl) {
      set_car($vl, $val);
      return 1;
   }
   else {
      $main::ots = $main::ots . 
                   sprintf("%s", "setVar: variable not defined: ");
      prName($nm);
      $main::ots = $main::ots . sprintf("%s", "\n");
      die "Abnormal Termination.";
   }
};

sub getVar  {
   my $nm   = shift;
   my $lenv = shift;
   my $vl  = bindingInEnv($nm, $lenv);

   if ($vl) {
      return car($vl);
   }
   else {
      return $falseValue;
   }
};

sub extendEnv {
   my $lenv = shift;
   my $lvars = shift;
   my $lvals = shift;

   return newENV($lvars, $lvals, $lenv);
}

sub isbound  {
   my $nm   = shift;
   my $lenv = shift;
   my $temp;
   $temp =  bindingInEnv($nm, $lenv);
   return ($temp != $nil);
};

sub newENV {
   my $type  = 'ENV';
   my $vars  = shift;
   my $vals  = shift;
   my $next  = shift;

   if (!pairp($vars) || !pairp($vals) || $next->('type') ne 'ENV' ) {
      $main::ots = $main::ots . 
                   sprintf("%s", "invalid arguments to newENV!");
      die "Abnormal Termination.";
   }
   my $objref     = sub {
                         my $msg = shift;
                         if ($msg eq 'getvars') {
                            return $vars;
                         } 
                         elsif ($msg eq 'getvals') {
                            return $vals;
                         } 
                         elsif ($msg eq 'setvars') {
                            my $lvars = shift;
                            $vars = $lvars;
                            return 1;
                         } 
                         elsif ($msg eq 'setvals') {
                            my $lvals = shift;
                            $vals = $lvals;
                            return 1;
                         } 
                         elsif ($msg eq 'getnext') {
                            return $next;
                         } 
                         elsif ($msg eq 'isempty') {
                            return ($vars == $nil && $next eq $null);
                         } 
                         elsif ($msg eq 'type') {
                            return $type;
                         } 
                         else {
                            $main::ots = $main::ots . 
                                         sprintf("%s", 
                                         "newENV: message unknown: $msg !\n");
                            die "Abnormal Termination.";
                         }
                    };
   return (bless $objref);
}

1;
